#include <stdexcept>
#include <regex>
#include <string>

#include "xml_codec_factory.hh"

#include "loggers.hh"
#include "xml_converters.hh"

#include "LibHttp_XmlMessageBodyTypes.hh"

int xml_codec::encode(const LibHttp__XmlMessageBodyTypes::XmlBody& body, OCTETSTRING& data)
{
  loggers::get_instance().log_msg(">>> xml_codec::encode: ", (const Base_Type&)body);

  if (body.raw().is_present() && body.raw().is_bound()) {
    const CHARSTRING& v = static_cast<const CHARSTRING&>(*body.raw().get_opt_value());
    data = char2oct(v);
  } else {
    TTCN_EncDec::clear_error();
    TTCN_EncDec::set_error_behavior(TTCN_EncDec::ET_ALL, TTCN_EncDec::EB_DEFAULT);
    TTCN_Buffer encoding_buffer;

    const LibHttp__XmlMessageBodyTypes::XmlBodyMsg& msg = body.msg();
    if (msg.ischosen(LibHttp__XmlMessageBodyTypes::XmlBodyMsg::ALT_raw)) { // If raw message is provided, msg is not encoded
      data = char2oct(msg.raw());
      loggers::get_instance().log_msg("xml_codec::encode: Process Raw: ", data);
      return 0;
    } else if (msg.ischosen(LibHttp__XmlMessageBodyTypes::XmlBodyMsg::ALT_pull__request)) {
      const http__www__cise__eu__servicemodel__v1__message::PullRequest& pull_request = msg.pull__request();
      loggers::get_instance().log_msg("xml_codec::encode: Process PullRequest: ", (const Base_Type&)pull_request);
      pull_request.encode(http__www__cise__eu__servicemodel__v1__message::PullRequest_descr_, encoding_buffer, TTCN_EncDec::CT_XER, XER_EXTENDED);
    } else if (msg.ischosen(LibHttp__XmlMessageBodyTypes::XmlBodyMsg::ALT_pull__response)) {
      const http__www__cise__eu__servicemodel__v1__message::PullResponse& pull_response = msg.pull__response();
      loggers::get_instance().log_msg("xml_codec::encode: Process PullResponse: ", (const Base_Type&)pull_response);
      pull_response.encode(http__www__cise__eu__servicemodel__v1__message::PullResponse_descr_, encoding_buffer, TTCN_EncDec::CT_XER, XER_EXTENDED);
    } else if (msg.ischosen(LibHttp__XmlMessageBodyTypes::XmlBodyMsg::ALT_acknowledgement)) {
      const http__www__cise__eu__servicemodel__v1__message::Acknowledgement& acknowledgement= msg.acknowledgement();
      loggers::get_instance().log_msg("xml_codec::encode: Process Acknowledgement: ", (const Base_Type&)acknowledgement);
      acknowledgement.encode(http__www__cise__eu__servicemodel__v1__message::Acknowledgement_descr_, encoding_buffer, TTCN_EncDec::CT_XER, XER_EXTENDED);
    } else if (msg.ischosen(LibHttp__XmlMessageBodyTypes::XmlBodyMsg::ALT_push)) {
      const http__www__cise__eu__servicemodel__v1__message::Push& push= msg.push();
      loggers::get_instance().log_msg("xml_codec::encode: Process Push: ", (const Base_Type&)push);
      push.encode(http__www__cise__eu__servicemodel__v1__message::Push_descr_, encoding_buffer, TTCN_EncDec::CT_XER, XER_EXTENDED);
    } else {
      loggers::get_instance().warning("xml_codec::encode: Unsupported variant");
      return -1;
    }

    // Canonicalization of the XML document
    loggers::get_instance().log("xml_codec::encode: Canonicalization of the XML document");
    const std::string to_canonical(static_cast<const char*>((const char*)encoding_buffer.get_data()), encoding_buffer.get_len() + static_cast<const char*>((const char*)encoding_buffer.get_data()));
    std::string canonicalized;
    if (xml_converters::get_instance().xml_canonicalization(to_canonical, canonicalized) == -1) {
      data = OCTETSTRING(encoding_buffer.get_len(), encoding_buffer.get_data());
    } else {
      data = OCTETSTRING(canonicalized.length(), (const unsigned char*)canonicalized.c_str());
    }

    { // TITAN issue: variant "namespace for 'agent/organization' seems buggy: <Agent xsi:type="agent:Organization">
      //        ==> xmlns:xsi=http://www.cise.eu/datamodel/v1/entity/organization/ xsi:type="ns4:Organization"
      loggers::get_instance().log("xml_codec::encode: Apply TITAN workaround");
      std::string s(static_cast<const unsigned char*>(data), data.lengthof() + static_cast<const unsigned char*>(data));
      patch_encode_agent_derivations(s);
      //loggers::get_instance().log("xml_codec::encode: After patch_encode_agent_derivations: '%s'", s.c_str());
      patch_encode_location_derivations(s);
      //loggers::get_instance().log("xml_codec::encode: After patch_encode_location_derivations: '%s'", s.c_str());
      patch_encode_event_derivations(s);
      //loggers::get_instance().log("xml_codec::encode: After patch_encode_event_derivations: '%s'", s.c_str());
      patch_encode_document_derivations(s);
      //loggers::get_instance().log("xml_codec::encode: After patch_encode_document_derivations: '%s'", s.c_str());
      patch_encode_service_capability_derivations(s);
      //loggers::get_instance().log("xml_codec::encode: After patch_encode_service_capability_derivations: '%s'", s.c_str());

      data = OCTETSTRING(s.length(), (const unsigned char*)s.c_str());
    }

    // Add header
    CHARSTRING h("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?>\n");
    data = char2oct(h) + data;
  }

  loggers::get_instance().log_msg("xml_codec::encode: After encoding: ", data);

  loggers::get_instance().log("<<< xml_codec::encode");
  return 0;
}

int xml_codec::patch_encode_agent_derivations(std::string& s) {
  std::string search;
  std::string replace;
  size_t idx = 0;
  if (s.find(" xsi:type=\"agent:Agent") != std::string::npos) {
    search.assign(" xsi:type=\"agent:Agent\"");
    replace.assign(" xsi:type=\"ns6:Agent\" xmlns:ns6=\"http://www.cise.eu/datamodel/v1/entity/organization/\"");
    //loggers::get_instance().log("xml_codec::patch_encode_agent_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
  }
  if (s.find(" xsi:type=\"agent:Organization") != std::string::npos) {
    idx = 0;
    search.assign(" xsi:type=\"agent:Organization\"");
    replace.assign(" xsi:type=\"ns6:Organization\" xmlns:ns6=\"http://www.cise.eu/datamodel/v1/entity/organization/\"");
    //loggers::get_instance().log("xml_codec::patch_encode_agent_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
  }
  if (s.find(" xsi:type=\"agent:FormalOrganization") != std::string::npos) {
    idx = 0;
    search.assign(" xsi:type=\"agent:FormalOrganization\"");
    replace.assign(" xsi:type=\"ns6:FormalOrganization\" xmlns:ns6=\"http://www.cise.eu/datamodel/v1/entity/organization/\"");
    //loggers::get_instance().log("xml_codec::patch_encode_agent_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
  }
  if (s.find(" xsi:type=\"agent:OrganizationalCollaboration") != std::string::npos) {
    idx = 0;
    search.assign(" xsi:type=\"agent:OrganizationalCollaboration\"");
    replace.assign(" xsi:type=\"ns6:OrganizationalCollaboration\" xmlns:ns6=\"http://www.cise.eu/datamodel/v1/entity/organization/\"");
    //loggers::get_instance().log("xml_codec::patch_encode_agent_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
  }
  if (s.find(" xsi:type=\"agent:OrganizationalUnit") != std::string::npos) {
    idx = 0;
    search.assign(" xsi:type=\"agent:OrganizationalUnit\"");
    replace.assign(" xsi:type=\"ns6:OrganizationalUnit\" xmlns:ns6=\"http://www.cise.eu/datamodel/v1/entity/organization/\"");
    //loggers::get_instance().log("xml_codec::patch_encode_agent_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
  }
  if (s.find(" xsi:type=\"agent:Person") != std::string::npos) {
    idx = 0;
    search.assign(" xsi:type=\"agent:Person\"");
    replace.assign(" xsi:type=\"ns6:Person\" xmlns:ns6=\"http://www.cise.eu/datamodel/v1/entity/organization/\"");
    //loggers::get_instance().log("xml_codec::patch_encode_agent_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
  }
  if (s.find(" xsi:type=\"agent:PortOrganization") != std::string::npos) {
    idx = 0;
    search.assign(" xsi:type=\"agent:PortOrganization\"");
    replace.assign(" xsi:type=\"ns6:PortOrganization\" xmlns:ns6=\"http://www.cise.eu/datamodel/v1/entity/organization/\"");
    //loggers::get_instance().log("xml_codec::patch_encode_agent_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
  }

  //loggers::get_instance().log("xml_codec::patch_encode_agent_derivations: s: '%s'", s.c_str());
  return 0;
}

int xml_codec::patch_encode_location_derivations(std::string& s) {
  std::string search;
  std::string replace;
  size_t idx = 0;
  if (s.find(" xsi:type=\"location:Location") != std::string::npos) {
    search.assign(" xsi:type=\"location:Location\"");
    replace.assign(" xsi:type=\"ns6:Location\" xmlns:ns6=\"http://www.cise.eu/datamodel/v1/entity/location/\"");
    //loggers::get_instance().log("xml_codec::patch_encode_location_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
  }
  if (s.find(" xsi:type=\"location:NamedLocation") != std::string::npos) {
    idx = 0;
    search.assign(" xsi:type=\"location:NamedLocation\"");
    replace.assign(" xsi:type=\"ns6:NamedLocation\" xmlns:ns6=\"http://www.cise.eu/datamodel/v1/entity/location/\"");
    //loggers::get_instance().log("xml_codec::patch_encode_location_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
  }
  if (s.find(" xsi:type=\"location:PortFacilityLocation") != std::string::npos) {
    idx = 0;
    search.assign(" xsi:type=\"location:PortFacilityLocation\"");
    replace.assign(" xsi:type=\"ns6:PortFacilityLocation\" xmlns:ns6=\"http://www.cise.eu/datamodel/v1/entity/location/\"");
    //loggers::get_instance().log("xml_codec::patch_encode_location_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
  }
  if (s.find(" xsi:type=\"location:PortLocation") != std::string::npos) {
    idx = 0;
    search.assign(" xsi:type=\"location:PortLocation\"");
    replace.assign(" xsi:type=\"ns6:PortLocation\" xmlns:ns6=\"http://www.cise.eu/datamodel/v1/entity/location/\"");
    //loggers::get_instance().log("xml_codec::patch_encode_location_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
  }

  //loggers::get_instance().log("xml_codec::patch_encode_location_derivations: s: '%s'", s.c_str());
  return 0;
}

int xml_codec::patch_encode_event_derivations(std::string& s) {
  std::string search;
  std::string replace;
  size_t idx = 0;
  if (s.find(" xsi:type=\"event:Event") != std::string::npos) {
    search.assign(" xsi:type=\"event:Event\"");
    replace.assign(" xsi:type=\"ns6:Event\" xmlns:ns6=\"http://www.cise.eu/datamodel/v1/entity/event/\"");
    //loggers::get_instance().log("xml_codec::patch_encode_event_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
  }

  if (s.find("<Event>") != std::string::npos) {
    idx = 0;
    search.assign("<Event>");
    replace.assign("<Event xsi:type=\"ns6:Event\" xmlns:ns6=\"http://www.cise.eu/datamodel/v1/entity/event/\">");
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
  }
  
  if (s.find(" xsi:type=\"action:Action") != std::string::npos) {
    idx = 0;
    search.assign(" xsi:type=\"action:Action\"");
    replace.assign(" xsi:type=\"ns6:Action\" xmlns:ns6=\"http://www.cise.eu/datamodel/v1/entity/action/\"");
    //loggers::get_instance().log("xml_codec::patch_encode_event_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
  }
  if (s.find(" xsi:type=\"anomaly:Anomaly") != std::string::npos) {
    idx = 0;
    search.assign(" xsi:type=\"anomaly:Anomaly\"");
    replace.assign(" xsi:type=\"ns6:Anomaly\" xmlns:ns6=\"http://www.cise.eu/datamodel/v1/entity/anomaly/\"");
    //loggers::get_instance().log("xml_codec::patch_encode_event_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
  }
  if (s.find(" xsi:type=\"incident:CrisisIncident") != std::string::npos) {
    idx = 0;
    search.assign(" xsi:type=\"incident:CrisisIncident\"");
    replace.assign(" xsi:type=\"ns6:CrisisIncident\" xmlns:ns6=\"http://www.cise.eu/datamodel/v1/entity/incident/\"");
    //loggers::get_instance().log("xml_codec::patch_encode_event_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
  }
  if (s.find(" xsi:type=\"incident:Incident") != std::string::npos) {
    idx = 0;
    search.assign(" xsi:type=\"incident:Incident\"");
    replace.assign(" xsi:type=\"ns6:Incident\" xmlns:ns6=\"http://www.cise.eu/datamodel/v1/entity/incident/\"");
    //loggers::get_instance().log("xml_codec::patch_encode_event_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
  }
  if (s.find(" xsi:type=\"incident:IrregularMigrationIncident") != std::string::npos) {
    idx = 0;
    search.assign(" xsi:type=\"incident:IrregularMigrationIncident\"");
    replace.assign(" xsi:type=\"ns6:IrregularMigrationIncident\" xmlns:ns6=\"http://www.cise.eu/datamodel/v1/entity/incident/\"");
    //loggers::get_instance().log("xml_codec::patch_encode_event_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
  }
  if (s.find(" xsi:type=\"incident:LawInfringementIncident") != std::string::npos) {
    idx = 0;
    search.assign(" xsi:type=\"incident:LawInfringementIncident\"");
    replace.assign(" xsi:type=\"ns6:LawInfringementIncident\" xmlns:ns6=\"http://www.cise.eu/datamodel/v1/entity/incident/\"");
    //loggers::get_instance().log("xml_codec::patch_encode_event_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
  }
  if (s.find(" xsi:type=\"incident:MaritimeSafetyIncident") != std::string::npos) {
    idx = 0;
    search.assign(" xsi:type=\"incident:MaritimeSafetyIncident\"");
    replace.assign(" xsi:type=\"ns6:MaritimeSafetyIncident\" xmlns:ns6=\"http://www.cise.eu/datamodel/v1/entity/incident/\"");
    //loggers::get_instance().log("xml_codec::patch_encode_event_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
  }
  if (s.find(" xsi:type=\"movement:Movement") != std::string::npos) {
    idx = 0;
    search.assign(" xsi:type=\"movement:Movement\"");
    replace.assign(" xsi:type=\"ns6:Movement\" xmlns:ns6=\"http://www.cise.eu/datamodel/v1/entity/movement/\"");
    //loggers::get_instance().log("xml_codec::patch_encode_event_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
  }
  if (s.find(" xsi:type=\"incident:PollutionIncident") != std::string::npos) {
    idx = 0;
    search.assign(" xsi:type=\"incident:PollutionIncident\"");
    replace.assign(" xsi:type=\"ns6:PollutionIncident\" xmlns:ns6=\"http://www.cise.eu/datamodel/v1/entity/incident/\"");
    //loggers::get_instance().log("xml_codec::patch_encode_event_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
  }

  //loggers::get_instance().log("xml_codec::patch_encode_event_derivations: s: '%s'", s.c_str());
  return 0;
}

int xml_codec::patch_encode_object_derivations(std::string& s) {

  return 0;
}

int xml_codec::patch_encode_document_derivations(std::string& s) {

  return 0;
}

int xml_codec::patch_encode_service_capability_derivations(std::string& s) {
  std::string search;
  std::string replace;
  size_t idx = 0;
  if (s.find("tq0000:SubscriptionCapability") != std::string::npos) {
    search.assign("tq0000:SubscriptionCapability");
    replace.assign("ns3:SubscriptionCapability");
    idx = 0;
    loggers::get_instance().log("xml_codec::patch_encode_service_capability_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    while ((idx = s.find(search, idx)) != std::string::npos) {
      s.replace(idx, search.length(), replace);
      idx += replace.length();
    } // End of 'while' statement
    search.assign("<ns4:PullRequest ");
    replace.assign("<ns4:PullRequest xmlns:ns3=\"http://www.cise.eu/servicemodel/v1/service/\" ");
    loggers::get_instance().log("xml_codec::patch_encode_service_capability_derivations: Apply TITAN workaround once: %s --> %s", search.c_str(), replace.c_str());
    idx = s.find(search);
    s.replace(idx, search.length(), replace);
    // search.assign(" xsi:type=\"tq0000:SubscriptionCapability\"");
    // replace.assign("");
    // idx = 0;
    // loggers::get_instance().log("xml_codec::patch_encode_service_capability_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    // while ((idx = s.find(search, idx)) != std::string::npos) {
    //   s.replace(idx, search.length(), replace);
    //   idx += replace.length();
    // } // End of 'while' statement
    // if (s.find("<Requests xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"") != std::string::npos) {
    //   search.assign("<Requests xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"");
    //   replace.assign("<Requests xsi:type=\"ns4:SubscriptionCapability\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"");
    //   loggers::get_instance().log("xml_codec::patch_encode_service_capability_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
    //   idx = 0;
    //   while ((idx = s.find(search, idx)) != std::string::npos) {
    //     s.replace(idx, search.length(), replace);
    //     idx += replace.length();
    //   } // End of 'while' statement
    // }
  }

  return 0;
}

int xml_codec::decode (const OCTETSTRING& p_data, LibHttp__XmlMessageBodyTypes::XmlBody& body, params* p_params)
{
  loggers::get_instance().log_msg(">>> xml_codec::decode: p_data=", p_data);

  // Sanity checks
  params::const_iterator it;
  if (p_params == nullptr) {
    loggers::get_instance().warning("xml_codec::decode: Failed to access p_params (null pointer)");
    return -1; // TODO Use p_data instead of return -1
  } else {
    it = p_params->find("decode_str");
    if (it == p_params->cend()) {
      loggers::get_instance().warning("xml_codec::decode: Failed to access p_params item (decode_str)");
      return -1; // TODO Use p_data instead of return -1
    }
    loggers::get_instance().log("xml_codec::decode: it->second='%s'", it->second.c_str());
  }

  // TITAN issue: variant "namespace for 'agent/organization' seems buggy: <Agent xsi:type="agent:Organization">
    //        ==> xmlns:xsi=http://www.cise.eu/datamodel/v1/entity/organization/ xsi:type="ns4:Organization"
    loggers::get_instance().log("xml_codec::decode: Apply TITAN workaround");
    std::string s(it->second);
    patch_decode_service_capability_derivations(s);
    //loggers::get_instance().log("xml_codec::decode: After patch_decode_service_capability_derivations: '%s'", s.c_str());


  loggers::get_instance().log_msg("xml_codec::decode: body: ", body);
  LibHttp__XmlMessageBodyTypes::XmlBodyMsg& msg = body.msg();
  body.raw() = OPTIONAL<CHARSTRING>(CHARSTRING(s.c_str()));
 TTCN_EncDec::set_error_behavior(TTCN_EncDec::ET_ALL, TTCN_EncDec::EB_DEFAULT);
  TTCN_EncDec::clear_error();
  TTCN_Buffer decoding_buffer(OCTETSTRING(s.length(), (const unsigned char*)s.c_str()));
 
  if ((s.find("<PullRequest") != std::string::npos) || (s.find(":PullRequest") != std::string::npos)) {
    http__www__cise__eu__servicemodel__v1__message::PullRequest pull_request;
    pull_request.decode(http__www__cise__eu__servicemodel__v1__message::PullRequest_descr_, decoding_buffer, TTCN_EncDec::CT_XER, XER_EXTENDED);
    msg.pull__request() = pull_request;
  } else if ((s.find("<PullResponse") != std::string::npos) || (s.find(":PullResponse") != std::string::npos)) {
    http__www__cise__eu__servicemodel__v1__message::PullResponse pull_response;
    pull_response.decode(http__www__cise__eu__servicemodel__v1__message::PullResponse_descr_, decoding_buffer, TTCN_EncDec::CT_XER, XER_EXTENDED);
    msg.pull__response() = pull_response;
  } else if ((s.find("<Acknowledgement") != std::string::npos) || (s.find(":Acknowledgement") != std::string::npos)) {
    http__www__cise__eu__servicemodel__v1__message::Acknowledgement acknowledgement;
    acknowledgement.decode(http__www__cise__eu__servicemodel__v1__message::Acknowledgement_descr_, decoding_buffer, TTCN_EncDec::CT_XER, XER_EXTENDED);
    msg.acknowledgement() = acknowledgement;
  } else if ((s.find("<Push") != std::string::npos) || (s.find(":Push") != std::string::npos)) {
    http__www__cise__eu__servicemodel__v1__message::Push push;
    push.decode(http__www__cise__eu__servicemodel__v1__message::Push_descr_, decoding_buffer, TTCN_EncDec::CT_XER, XER_EXTENDED);
    msg.push() = push;
  } else {
    loggers::get_instance().warning("xml_codec::decode: Unsupported variant");
    return -1;
  }

  loggers::get_instance().log_msg("<<< xml_codec::decode: ", (const Base_Type&)msg);
  return 0;
}

int xml_codec::patch_decode_service_capability_derivations(std::string& s) {

    std::string search;
    std::string replace;
    size_t idx = 0;
    if (s.find("<ServiceCapability/>") != std::string::npos) {
      search.assign("<ServiceCapability/>");
      //loggers::get_instance().log("xml_codec::patch_decode_service_capability_derivations: Apply TITAN workaround: %s --> %s", search.c_str(), replace.c_str());
      while ((idx = s.find(search, idx)) != std::string::npos) {
        s = s.erase(idx, search.length());
        idx += 1;
      } // End of 'while' statement
    }

  return 0;
}

xml_codec_factory xml_codec_factory::_f;
